from math import ceil
from typing import Callable

def raise_type_error(var,name,types):
	if type(types) is not tuple: types = (types,)
	if var and not isinstance(var,types): raise TypeError(f"{name} must be {' or '.join([x.__name__ for x in types])}")

class State:

	def __init__(self,name,msg,type=int,parse_mode=None,back=False,webpage_preview=False):
		self.name = name
		self.data_type = type
		self.msg = msg
		self.mode = parse_mode
		self.callback = None
		self.text = None
		self.action = None
		self.build = None
		self.back = back
		self.webpage_preview = webpage_preview
		self.routes = None
	
	def __str__(self):
		return f'State <{self.name}>'

	def add_keyboard(self,keyboard,size=None,max_row=3):
		'''Add inline keyboard for the state'''
		raise_type_error(keyboard,'keyboard',(list,dict))
		raise_type_error(size,'size',tuple)
		if isinstance(keyboard,list): keyboard = dict(enumerate(keyboard))
		if size and sum(size) != len(keyboard): raise ValueError(f'Keyboard length ({len(keyboard)}) must be the same size as the sum of row size ({sum(size)}).')
		elem_n = len(keyboard)
		if not size: size = [max_row for _ in range(elem_n//max_row)] + ([r,] if (r := elem_n % max_row) else [])
		self.callback = (keyboard,tuple(size))

	def add_text(self,regex=None,error=None):
		'''Add handler for text input'''
		raise_type_error(regex,'regex',str)
		raise_type_error(error,'error',str)
		if not regex: regex = r'^.*$' 
		self.text = (regex,error)

	def add_action(self,function):
		'''Add action to do with this command'''
		raise_type_error(function,'function',(Callable))
		self.action = function

	def add_dynamic_keyboard(self,function,max_row=3):
		'''Add action to build a keyboard dynamically'''
		raise_type_error(function,'function',(Callable))
		raise_type_error(max_row,'max_row',(int))
		self.build = function
		self.max_row = max_row

	def add_dynamic_routes(self,function):
		'''Add function to create dynamic routes | must return 3 value (routes,default,back)'''
		raise_type_error(function,'function',(Callable))
		self.routes = function



from autoconv.state import State

def raise_type_error(var,name,types):
	if type(types) is not tuple: types = (types,)
	if var and not isinstance(var,types): raise TypeError(f"{name} must be {' or '.join([x.__name__ for x in types])}")

class Conversation:

	def __init__(self,start_state,end_state=None):
		self.start = start_state
		self.end = end_state
		self.state_list = [start_state,end_state] if end_state else [start_state]
		self.routes = {}

	def __str__(self):
		routes_print = '\n'.join(['  * {}:\n{}'.format(s,'\n'.join(['      {:>5} -> {}'.format(v,d) for v,d in r.items()])) for s,r in self.routes.items()])
		heading = f'CONVERSATION\n{self.start} ==> ' + (f'{self.end}' if self.end else f'...')
		return f'{heading}\n# State list: {[str(s) for s in self.state_list]}\n# Routes:\n{routes_print}'

	def _state_already_exists(self,state):
		'''Check for duplicates'''
		return any([state.name == s.name for s in self.state_list])

	def add_state(self,state):
		'''Add state to the conversation'''
		raise_type_error(state,'state',(State,list))
		for s in state:
			raise_type_error(s,'state',State)
			if not self._state_already_exists(s): self.state_list.append(s)
			else: raise ValueError(f'Already exists a state with name <{s.name}>.') 

	def add_routes(self,state,routes=None,default=None,back=None):
		'''Add state routes'''
		raise_type_error(state,'state',State)
		raise_type_error(routes,'routes',dict)
		raise_type_error(default,'default',State)
		raise_type_error(back,'back',State)
		if state not in self.state_list: raise ValueError(f'{state} doesn\'t exist in this conversation.')
		if routes:
			for k,v in routes.items():
				if (state.callback and k not in state.callback[0]) and k in (-1,'BACK'): raise ValueError(f'\'{k}\' it\'s not a possible value of {str(state)}.')
				if v not in self.state_list: raise ValueError(f'{str(v)} doesn\'t exist in this conversation.')
			self.routes.update({state.name:routes})
		if default:
			if default not in self.state_list: raise ValueError(f'{str(default)} doesn\'t exist in this conversation.')
			s.update({-1:default}) if (s := self.routes.get(state.name)) else self.routes.update({state.name:{-1:default}})
		if back:
			if back not in self.state_list: raise ValueError(f'{str(default)} doesn\'t exist in this conversation.')
			s.update({'BACK':back}) if (s := self.routes.get(state.name)) else self.routes.update({state.name:{'BACK':back}})

	def get_state(self,state_name):
		'''Get state from states list by name'''
		for state in self.state_list:
			if state.name == state_name: return state
		return None